//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gertespezifischer Funktionen und Mechanismen            //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include "AVRChip.h"
#include "Conf.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum
{   DO_CONTINUE,
    DO_RESTART,
    DO_ENTER_APPL
} DoNext;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt16           prgLdrMode __attribute__ ((section (".noinit")));
static volatile UInt16  msecTicks                                  =   0;
static volatile UInt08  busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY+1]    = { 0 };
static volatile UInt08  busRxQueueHead                             =   0;
static volatile UInt08  busRxQueueTail                             =   0;
static volatile UInt08  serRxQueue[CFG_SERIAL_RX_QUEUE_CAPACITY+1] = { 0 };
static volatile UInt08  serRxQueueHead                             =   0;
static volatile UInt08  serRxQueueTail                             =   0;
static          DoNext  doNext                                     = DO_CONTINUE;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void __attribute__((noreturn))
Restart(void)
{
    DisableInterrupts();
    prgLdrMode = CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while( true );
}

//////////////////////////////////////////////////////////////////////////////////////////
static void __attribute__((noreturn))
EnterAppl(void)
{
    DisableInterrupts();
    prgLdrMode = !CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while( true );
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Vorlauf  /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                           _                                                          //
//                          (_)                                                         //
//                           |                                                          //
//                           |                                                          //
//                          \|/                                                         //
//                    /\  ja                                              //
//                   | Power-On-Reset |------------------+                              //
//                    \______________/                   |                              //
//                           | nein                      |                              //
//                           |                           |                              //
//                          \|/                         \|/                             //
//                    /\ nein       /\ ja                   //
//                   | External-Reset |--------->| Watchdog-Reset |------+              //
//                    \______________/            \______________/       |              //
//                           | ja                        | nein          |              //
//                           |                           |               |              //
//                          \|/                         \|/              |              //
//                +----------------------+   +-----------------------+   |              //
//                | mode = PGMLDR_MARKER |   | mode = !PGMLDR_MARKER |   |              //
//                +----------------------+   +-----------------------+   |              //
//                           |                           |               |              //
//                           |<--------------------------+---------------+              //
//                           |                                                          //
//                          \|/                                                         //
//                /\  nein                                         //
//               | mode != PGMLDR_MARKER |-----------------+                            //
//                \_____________________/                  |                            //
//                           | ja                          |                            //
//                           |                             |                            //
//                          \|/                            |                            //
//                  /\  nein              |                            //
//                 | Read(0) != 0xFFFF |------------------>|                            //
//                  \_________________/                    |                            //
//                           |  ja                         |                            //
//                           |                             |                            //
//                   ________|________            _________|__________                  //
//                   Enter Application            Enter Program Loader                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked)) __attribute__((section(".init3"))) void
PreStart(void)
{
    UInt08 mcusr = MCUSR;
    MCUSR = 0;
    UInt16 mode = prgLdrMode;
    wdt_enable(WDTO_2S);
    if( mcusr & BV(PORF) )
    {   mode = !CFG_PGMLDR_MARKER;
    }
    else
    {   if( mcusr & BV(EXTRF) )
        {   mode = CFG_PGMLDR_MARKER;
        }
        else
        {   if( !(mcusr & BV(WDRF)) )
            {   mode = !CFG_PGMLDR_MARKER;
            }
        }
    }
    if (mode != CFG_PGMLDR_MARKER && pgm_read_word_near(0) != 0xFFFF) GOTO(0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_RX_vect)
{
    UInt08 byte = UDR0;
    UInt08 tail = busRxQueueTail+1;
    if( tail == CARDINALITY(busRxQueue) ) tail = 0;
    if( tail == busRxQueueHead ) return;
    busRxQueue[busRxQueueTail] = byte;
    busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART1_RX_vect)
{
    UInt08 status = UCSR1A;
    UInt08 byte   = UDR1;
    if( status & (BV(FE1)|BV(DOR1)) ) return;
    UInt08 tail = serRxQueueTail+1;
    if( tail == CARDINALITY(serRxQueue) ) tail = 0;
    if( tail == serRxQueueHead ) return;
    serRxQueue[serRxQueueTail] = byte;
    serRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(__vector_default) __attribute__((naked));
ISR(__vector_default)
{
    // Programmlader Einstiegspunkt
    Restart();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Restart(void)
{
    doNext = DO_RESTART;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterAppl(void)
{
    doNext = DO_ENTER_APPL;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceiveByte(UInt08* byte)
{
    if( busRxQueueHead == busRxQueueTail )
    {   *byte = '\0';
        return false;
    }
    Atomic
    {   *byte = busRxQueue[busRxQueueHead];
        busRxQueueHead++;
        if( busRxQueueHead == CARDINALITY(busRxQueue) ) busRxQueueHead = 0;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SerTransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR1A, UDRE1);
    UDR1 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_SerReceiveByte(UInt08* byte)
{
    if( serRxQueueHead == serRxQueueTail )
    {   *byte = '\0';
        return false;
    }
    Atomic
    {   *byte = serRxQueue[serRxQueueHead];
        serRxQueueHead++;
        if( serRxQueueHead == CARDINALITY(serRxQueue) ) serRxQueueHead = 0;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(UInt16 adr, void* mem, UInt16 size)
{
    eeprom_read_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(UInt16 adr, const void* mem, UInt16 size)
{
    eeprom_update_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
	// Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(A, _P_,    // PA.0: A0_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.1: A1_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.2: A2_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.3: A3_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.4: A4_PDN	    : Allgemeiner Pull-Down-Anschluss
                _T_,    // PA.5: A5_PUP     : Allgemeiner Pull-Up-Anschluss
                _T_,    // PA.6: A6_AIN     : Analogeingang 0 .. 5V
                _L_);   // PA.7: A7_SYS_LED : System-Led: L = aus
    ConfPort(B, _P_,    // PB.0: B0_GPIO    : Allgemeiner Anschluss
                _T_,    // PB.1: B1_TA1     : Allgemeine Drucktaste 1, L = gedrckt
                _T_,    // PB.2: B2_TA2     : Allgemeine Drucktaste 2  L = gedrckt
                _L_,    // PB.3: B3_PIEZO   : Piezosignalgeber, Frequenz ber Timer 0 PWM
                _P_,    // PB.4: B4_GPIO    : Allgemeiner Anschluss
                _P_,    // PB.5: B5_MOSI    : SPI/ISP
                _P_,    // PB.6: B6_MISO    : SPI/ISP
                _P_);   // PB.7: B7_SCK     : SPI/ISP
    ConfPort(C, _P_,    // PC.0: C0_SCL     : TWI (IC)
                _P_,    // PC.1: C1_SDA     : TWI (IC)
                _P_,    // PC.2: C2_TCK     : JTAG
                _P_,    // PC.3: C3_TMS     : JTAG
                _P_,    // PC.4: C4_TDO     : JTAG
                _P_,    // PC.5: C5_TDI     : JTAG
                _L_,    // PC.6: C6_BUS_ACT : Cy4NET-Aktivitt, L = keine
                _L_);   // PC.7: C7_SER_ACT : Serielle-Aktivitt, L = keine
    ConfPort(D, _P_,    // PD.0: RXD0       : Bus-RXD
                _P_,    // PD.1: TXD0       : Bus-TXD
                _P_,    // PD.2: RXD1       : RS232-RXD
                _P_,    // PD.3: TXD1       : RS232-TXD
                _P_,    // PD.4: D4_GPIO    : Allgemeiner Anschluss
                _L_,    // PD.5: D5_GP_LED3 : Allgemeine LED 3, H = Ein
                _L_,    // PD.6: D6_GP_LED2 : Allgemeine LED 2, H = Ein
                _L_);   // PD.7: D7_GP_LED1 : Allgemeine LED 1, H = Ein (PWM Timer 2)
    // Interrupt-Vektortabelle in den Bootloader-Bereich verlegen
    asm volatile(    "push  R22      \n\t"
                     "IN    R22, %0  \n\t"
                     "SBR   R22, %1  \n\t"
                     "OUT   %0, R22; \n\t"
                     "CBR   R22, %1  \n\t"
                     "SBR   R22, %2  \n\t"
                     "OUT   %0, r22; \n\t"
                     "POP   R22      \n\t"
                    :
                    :"M" (_SFR_IO_ADDR(MCUCR)), "i" (BV(IVCE)), "i" (BV(IVSEL))
                    :"r22");
    Freeze100us();
    // 16-Bit Timer 1 ohne IRQ zhlen lassen (F_CPU1024) --------------------------------
    TCCR1B = BV(CS12) | BV(CS10);
    // USART 0 19200 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC ------------------
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(RXEN0) | BV(TXEN0);
    UBRR0  = F_CPU/16L/19200-1;
    // USART 1 57600 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC ------------------
    UCSR1A = 0;
    UCSR1B = BV(RXCIE1) | BV(RXEN1) | BV(TXEN1);
    UBRR1  = F_CPU/16L/57600-1;
	// Systemzustandsleuchte konfigurieren
    ConfPin(CFG_SYSTEM_LED, _L_);
	// -----------------------------------------------------------------------------------
    EnableInterrupts();
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_KeepRunning(void)
{
    return doNext == DO_CONTINUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    wdt_reset();
    UInt16 ticks = TIMERTICKS/4;
	#if CFG_LED_LOW_ACTIVE == true
		ChgPin(CFG_SYSTEM_LED,  (ticks & 0x0640));
	#else
		ChgPin(CFG_SYSTEM_LED, !(ticks & 0x0640));
	#endif
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    switch( doNext )
    {   case DO_RESTART:    Restart();
        case DO_ENTER_APPL: EnterAppl();
        default:            doNext = DO_RESTART;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
